pragma solidity 0.5.9;

// Since the source `ERC677BridgeTokenRewardable` requires solc v0.4.24 but truffle
// doesn't allow using different versions of compiler at the same time, this flat
// source file for `ERC677BridgeTokenRewardable` was taken from
// https://github.com/poanetwork/poa-bridge-contracts/tree/8a7cf85bfbf47f0e2ab6c5b2149873a7e48dc206
// and adapted for solc v0.5.9.

// File: contracts/ERC677Receiver.sol

contract ERC677Receiver {
  function onTokenTransfer(address _from, uint _value, bytes calldata _data) external returns(bool);
}

// File: openzeppelin-solidity/contracts/token/ERC20/ERC20Basic.sol

/**
 * @title ERC20Basic
 * @dev Simpler version of ERC20 interface
 * @dev see https://github.com/ethereum/EIPs/issues/179
 */
contract ERC20Basic {
  event Transfer(address indexed from, address indexed to, uint256 value);
  function totalSupply() public view returns(uint256);
  function balanceOf(address who) public view returns(uint256);
  function transfer(address to, uint256 value) public returns(bool);
}

// File: openzeppelin-solidity/contracts/token/ERC20/ERC20.sol

/**
 * @title ERC20 interface
 * @dev see https://github.com/ethereum/EIPs/issues/20
 */
contract ERC20 is ERC20Basic {
  event Approval(address indexed owner, address indexed spender, uint256 value);
  function allowance(address owner, address spender) public view returns(uint256);
  function transferFrom(address from, address to, uint256 value) public returns(bool);
  function approve(address spender, uint256 value) public returns(bool);
}

// File: contracts/ERC677.sol

contract ERC677 is ERC20 {
  event Transfer(address indexed from, address indexed to, uint value, bytes data);
  function transferAndCall(address, uint, bytes calldata) external returns(bool);
}

// File: contracts/IBurnableMintableERC677Token.sol

contract IBurnableMintableERC677Token is ERC677 {
  function mint(address, uint256) public returns(bool);
  function burn(uint256 _value) public;
  function claimTokens(address _token, address payable _to) public;
}

// File: openzeppelin-solidity/contracts/math/SafeMath.sol

/**
 * @title SafeMath
 * @dev Math operations with safety checks that throw on error
 */
library SafeMath {

  /**
  * @dev Multiplies two numbers, throws on overflow.
  */
  function mul(uint256 a, uint256 b) internal pure returns (uint256 c) {
    // Gas optimization: this is cheaper than asserting 'a' not being zero, but the
    // benefit is lost if 'b' is also tested.
    // See: https://github.com/OpenZeppelin/openzeppelin-solidity/pull/522
    if (a == 0) {
      return 0;
    }

    c = a * b;
    assert(c / a == b);
    return c;
  }

  /**
  * @dev Integer division of two numbers, truncating the quotient.
  */
  function div(uint256 a, uint256 b) internal pure returns (uint256) {
    // assert(b > 0); // Solidity automatically throws when dividing by 0
    // uint256 c = a / b;
    // assert(a == b * c + a % b); // There is no case in which this doesn't hold
    return a / b;
  }

  /**
  * @dev Subtracts two numbers, throws on overflow (i.e. if subtrahend is greater than minuend).
  */
  function sub(uint256 a, uint256 b) internal pure returns (uint256) {
    assert(b <= a);
    return a - b;
  }

  /**
  * @dev Adds two numbers, throws on overflow.
  */
  function add(uint256 a, uint256 b) internal pure returns (uint256 c) {
    c = a + b;
    assert(c >= a);
    return c;
  }
}

// File: openzeppelin-solidity/contracts/token/ERC20/BasicToken.sol

/**
 * @title Basic token
 * @dev Basic version of StandardToken, with no allowances.
 */
contract BasicToken is ERC20Basic {
  using SafeMath for uint256;

  mapping(address => uint256) balances;

  uint256 totalSupply_;

  /**
  * @dev total number of tokens in existence
  */
  function totalSupply() public view returns (uint256) {
    return totalSupply_;
  }

  /**
  * @dev transfer token for a specified address
  * @param _to The address to transfer to.
  * @param _value The amount to be transferred.
  */
  function transfer(address _to, uint256 _value) public returns (bool) {
    require(_to != address(0));
    require(_value <= balances[msg.sender]);

    balances[msg.sender] = balances[msg.sender].sub(_value);
    balances[_to] = balances[_to].add(_value);
    emit Transfer(msg.sender, _to, _value);
    return true;
  }

  /**
  * @dev Gets the balance of the specified address.
  * @param _owner The address to query the the balance of.
  * @return An uint256 representing the amount owned by the passed address.
  */
  function balanceOf(address _owner) public view returns (uint256) {
    return balances[_owner];
  }

}

// File: openzeppelin-solidity/contracts/token/ERC20/BurnableToken.sol

/**
 * @title Burnable Token
 * @dev Token that can be irreversibly burned (destroyed).
 */
contract BurnableToken is BasicToken {

  event Burn(address indexed burner, uint256 value);

  /**
   * @dev Burns a specific amount of tokens.
   * @param _value The amount of token to be burned.
   */
  function burn(uint256 _value) public {
    _burn(msg.sender, _value);
  }

  function _burn(address _who, uint256 _value) internal {
    require(_value <= balances[_who]);
    // no need to require value <= totalSupply, since that would imply the
    // sender's balance is greater than the totalSupply, which *should* be an assertion failure

    balances[_who] = balances[_who].sub(_value);
    totalSupply_ = totalSupply_.sub(_value);
    emit Burn(_who, _value);
    emit Transfer(_who, address(0), _value);
  }
}

// File: openzeppelin-solidity/contracts/token/ERC20/DetailedERC20.sol

/**
 * @title DetailedERC20 token
 * @dev The decimals are only for visualization purposes.
 * All the operations are done using the smallest and indivisible token unit,
 * just as on Ethereum all the operations are done in wei.
 */
contract DetailedERC20 is ERC20 {
  string public name;
  string public symbol;
  uint8 public decimals;

  constructor(string memory _name, string memory _symbol, uint8 _decimals) public {
    name = _name;
    symbol = _symbol;
    decimals = _decimals;
  }
}

// File: openzeppelin-solidity/contracts/ownership/Ownable.sol

/**
 * @title Ownable
 * @dev The Ownable contract has an owner address, and provides basic authorization control
 * functions, this simplifies the implementation of "user permissions".
 */
contract Ownable {
  address public owner;

  event OwnershipRenounced(address indexed previousOwner);
  event OwnershipTransferred(
    address indexed previousOwner,
    address indexed newOwner
  );

  /**
   * @dev The Ownable constructor sets the original `owner` of the contract to the sender
   * account.
   */
  constructor() public {
    owner = msg.sender;
  }

  /**
   * @dev Throws if called by any account other than the owner.
   */
  modifier onlyOwner() {
    require(msg.sender == owner);
    _;
  }

  /**
   * @dev Allows the current owner to relinquish control of the contract.
   */
  function renounceOwnership() public onlyOwner {
    emit OwnershipRenounced(owner);
    owner = address(0);
  }

  /**
   * @dev Allows the current owner to transfer control of the contract to a newOwner.
   * @param _newOwner The address to transfer ownership to.
   */
  function transferOwnership(address _newOwner) public onlyOwner {
    _transferOwnership(_newOwner);
  }

  /**
   * @dev Transfers control of the contract to a newOwner.
   * @param _newOwner The address to transfer ownership to.
   */
  function _transferOwnership(address _newOwner) internal {
    require(_newOwner != address(0));
    emit OwnershipTransferred(owner, _newOwner);
    owner = _newOwner;
  }
}

// File: openzeppelin-solidity/contracts/token/ERC20/StandardToken.sol

/**
 * @title Standard ERC20 token
 *
 * @dev Implementation of the basic standard token.
 * @dev https://github.com/ethereum/EIPs/issues/20
 * @dev Based on code by FirstBlood: https://github.com/Firstbloodio/token/blob/master/smart_contract/FirstBloodToken.sol
 */
contract StandardToken is ERC20, BasicToken {

  mapping (address => mapping (address => uint256)) internal allowed;

  /**
   * @dev Transfer tokens from one address to another
   * @param _from address The address which you want to send tokens from
   * @param _to address The address which you want to transfer to
   * @param _value uint256 the amount of tokens to be transferred
   */
  function transferFrom(
    address _from,
    address _to,
    uint256 _value
  )
    public
    returns (bool)
  {
    require(_to != address(0));
    require(_value <= balances[_from]);
    require(_value <= allowed[_from][msg.sender]);

    balances[_from] = balances[_from].sub(_value);
    balances[_to] = balances[_to].add(_value);
    allowed[_from][msg.sender] = allowed[_from][msg.sender].sub(_value);
    emit Transfer(_from, _to, _value);
    return true;
  }

  /**
   * @dev Approve the passed address to spend the specified amount of tokens on behalf of msg.sender.
   *
   * Beware that changing an allowance with this method brings the risk that someone may use both the old
   * and the new allowance by unfortunate transaction ordering. One possible solution to mitigate this
   * race condition is to first reduce the spender's allowance to 0 and set the desired value afterwards:
   * https://github.com/ethereum/EIPs/issues/20#issuecomment-263524729
   * @param _spender The address which will spend the funds.
   * @param _value The amount of tokens to be spent.
   */
  function approve(address _spender, uint256 _value) public returns (bool) {
    allowed[msg.sender][_spender] = _value;
    emit Approval(msg.sender, _spender, _value);
    return true;
  }

  /**
   * @dev Function to check the amount of tokens that an owner allowed to a spender.
   * @param _owner address The address which owns the funds.
   * @param _spender address The address which will spend the funds.
   * @return A uint256 specifying the amount of tokens still available for the spender.
   */
  function allowance(
    address _owner,
    address _spender
   )
    public
    view
    returns (uint256)
  {
    return allowed[_owner][_spender];
  }

  /**
   * @dev Increase the amount of tokens that an owner allowed to a spender.
   *
   * approve should be called when allowed[_spender] == 0. To increment
   * allowed value is better to use this function to avoid 2 calls (and wait until
   * the first transaction is mined)
   * From MonolithDAO Token.sol
   * @param _spender The address which will spend the funds.
   * @param _addedValue The amount of tokens to increase the allowance by.
   */
  function increaseApproval(
    address _spender,
    uint _addedValue
  )
    public
    returns (bool)
  {
    allowed[msg.sender][_spender] = allowed[msg.sender][_spender].add(_addedValue);
    emit Approval(msg.sender, _spender, allowed[msg.sender][_spender]);
    return true;
  }

  /**
   * @dev Decrease the amount of tokens that an owner allowed to a spender.
   *
   * approve should be called when allowed[_spender] == 0. To decrement
   * allowed value is better to use this function to avoid 2 calls (and wait until
   * the first transaction is mined)
   * From MonolithDAO Token.sol
   * @param _spender The address which will spend the funds.
   * @param _subtractedValue The amount of tokens to decrease the allowance by.
   */
  function decreaseApproval(
    address _spender,
    uint _subtractedValue
  )
    public
    returns (bool)
  {
    uint oldValue = allowed[msg.sender][_spender];
    if (_subtractedValue > oldValue) {
      allowed[msg.sender][_spender] = 0;
    } else {
      allowed[msg.sender][_spender] = oldValue.sub(_subtractedValue);
    }
    emit Approval(msg.sender, _spender, allowed[msg.sender][_spender]);
    return true;
  }

}

// File: openzeppelin-solidity/contracts/token/ERC20/MintableToken.sol

/**
 * @title Mintable token
 * @dev Simple ERC20 Token example, with mintable token creation
 * @dev Issue: * https://github.com/OpenZeppelin/openzeppelin-solidity/issues/120
 * Based on code by TokenMarketNet: https://github.com/TokenMarketNet/ico/blob/master/contracts/MintableToken.sol
 */
contract MintableToken is StandardToken, Ownable {
  event Mint(address indexed to, uint256 amount);
  event MintFinished();

  bool public mintingFinished = false;

  modifier canMint() {
    require(!mintingFinished);
    _;
  }

  modifier hasMintPermission() {
    require(msg.sender == owner);
    _;
  }

  /**
   * @dev Function to mint tokens
   * @param _to The address that will receive the minted tokens.
   * @param _amount The amount of tokens to mint.
   * @return A boolean that indicates if the operation was successful.
   */
  function mint(
    address _to,
    uint256 _amount
  )
    hasMintPermission
    canMint
    public
    returns (bool)
  {
    totalSupply_ = totalSupply_.add(_amount);
    balances[_to] = balances[_to].add(_amount);
    emit Mint(_to, _amount);
    emit Transfer(address(0), _to, _amount);
    return true;
  }

  /**
   * @dev Function to stop minting new tokens.
   * @return True if the operation was successful.
   */
  function finishMinting() onlyOwner canMint public returns (bool) {
    mintingFinished = true;
    emit MintFinished();
    return true;
  }
}

// File: contracts/ERC677BridgeToken.sol

contract ERC677BridgeToken is
  IBurnableMintableERC677Token,
  DetailedERC20,
  BurnableToken,
  MintableToken
{
  address public bridgeContract;

  event ContractFallbackCallFailed(address from, address to, uint value);

  constructor(
    string memory _name,
    string memory _symbol,
    uint8 _decimals
  ) public DetailedERC20(
    _name,
    _symbol,
    _decimals
  ) {}

  function setBridgeContract(address _bridgeContract) onlyOwner public {
    require(_bridgeContract != address(0) && isContract(_bridgeContract));
    bridgeContract = _bridgeContract;
  }

  modifier validRecipient(address _recipient) {
    require(_recipient != address(0) && _recipient != address(this));
    _;
  }

  function transferAndCall(
    address _to,
    uint _value,
    bytes calldata _data
  )
    external
    validRecipient(_to)
    returns(bool)
  {
    require(superTransfer(_to, _value));
    emit Transfer(msg.sender, _to, _value, _data);

    if (isContract(_to)) {
      require(contractFallback(_to, _value, _data));
    }
    return true;
  }

  function getTokenInterfacesVersion() public pure returns(uint64 major, uint64 minor, uint64 patch) {
    return (2, 0, 0);
  }

  function superTransfer(address _to, uint256 _value) internal returns(bool)
  {
    return super.transfer(_to, _value);
  }

  function transfer(address _to, uint256 _value) public returns (bool)
  {
    require(superTransfer(_to, _value));
    if (isContract(_to) && !contractFallback(_to, _value, new bytes(0))) {
      if (_to == bridgeContract) {
        revert();
      } else {
        emit ContractFallbackCallFailed(msg.sender, _to, _value);
      }
    }
    return true;
  }

  function contractFallback(
    address _to,
    uint _value,
    bytes memory _data
  )
    private
    returns(bool)
  {
    (bool success,) = _to.call(
      abi.encodeWithSignature("onTokenTransfer(address,uint256,bytes)", msg.sender, _value, _data)
    );
    return success;
  }

  function isContract(address _addr)
    internal
    view
    returns (bool)
  {
    uint length;
    assembly { length := extcodesize(_addr) }
    return length > 0;
  }

  function finishMinting() public returns (bool) {
    revert();
  }

  function renounceOwnership() public onlyOwner {
    revert();
  }

  function claimTokens(address _token, address payable _to) public onlyOwner {
    require(_to != address(0));
    if (_token == address(0)) {
      _to.transfer(address(this).balance);
      return;
    }

    DetailedERC20 token = DetailedERC20(_token);
    uint256 balance = token.balanceOf(address(this));
    require(token.transfer(_to, balance));
  }

}

// File: contracts/ERC677BridgeTokenRewardable.sol

contract ERC677BridgeTokenRewardable is ERC677BridgeToken {

  address public blockRewardContract;
  address public stakingContract;

  constructor(
    string memory _name,
    string memory _symbol,
    uint8 _decimals
  ) public ERC677BridgeToken(
    _name,
    _symbol,
    _decimals
  ) {}

  modifier onlyBlockRewardContract() {
    require(msg.sender == blockRewardContract);
    _;
  }

  modifier onlyStakingContract() {
    require(msg.sender == stakingContract);
    _;
  }

  function setBlockRewardContract(address _blockRewardContract) onlyOwner public {
    require(_blockRewardContract != address(0) && isContract(_blockRewardContract));
    blockRewardContract = _blockRewardContract;
  }

  function setStakingContract(address _stakingContract) onlyOwner public {
    require(_stakingContract != address(0) && isContract(_stakingContract));
    require(balanceOf(_stakingContract) == 0);
    stakingContract = _stakingContract;
  }

  function mintReward(address[] calldata _receivers, uint256[] calldata _rewards) external onlyBlockRewardContract {
    uint256 receiversLength = _receivers.length;
    for (uint256 i = 0; i < receiversLength; i++) {
      uint256 amount = _rewards[i];

      if (amount == 0) continue;

      address to = _receivers[i];

      // Mint `amount` for `to`
      totalSupply_ = totalSupply_.add(amount);
      balances[to] = balances[to].add(amount);
      emit Mint(to, amount);
      emit Transfer(address(0), to, amount);
    }
  }

  function stake(address _staker, uint256 _amount) external onlyStakingContract {
    // Transfer `_amount` from `_staker` to `stakingContract`
    balances[_staker] = balances[_staker].sub(_amount);
    balances[stakingContract] = balances[stakingContract].add(_amount);
    emit Transfer(_staker, stakingContract, _amount);
  }

  function transfer(address _to, uint256 _value) public returns(bool) {
    require(_to != stakingContract);
    return super.transfer(_to, _value);
  }

  function transferFrom(address _from, address _to, uint256 _value) public returns(bool) {
    require(_to != stakingContract);
    return super.transferFrom(_from, _to, _value);
  }

}

contract ERC677BridgeTokenRewardableMock is ERC677BridgeTokenRewardable {
  constructor(
    string memory _name,
    string memory _symbol,
    uint8 _decimals
  ) public ERC677BridgeTokenRewardable(
    _name,
    _symbol,
    _decimals
  ) {
  }
}
